'' This object provides an I2C EEPROM read/write routine that can handle both 100KHz and
'' 400KHz bus speeds and EEPROM page sizes of 64, 128, or 256 bytes (or no paging/no delay
'' as with Ramtron serial RAM).  The SPIN interpreter can be started after reading, either
'' in the same COG used by these routines or in a free COG.  The control information is
'' passed in a 2 long parameter block whose address is passed to the COG when it is started.
'' The parameter block is updated when the operation is completed.  Note that these are shown
'' here as they appear in a long value rather than the order of the bytes in memory.

'' -------------------------------------------------------------------
'' |   cmd/status   |          I/O pin / device / address            |
'' -------------------------------------------------------------------
'' |           byte count           |          HUB address           |
'' -------------------------------------------------------------------

'' The EEPROM address is in the same format used by other routines with the I/O pin pair
'' in bits 21..19, the device address in bits 18..16, and the 64K address in bits 15..0.
'' Note that the I/O pin pair is the number of the SCL pin divided by 2.  The SDA pin is
'' always the next higher numbered pin.  The command code is in the low order bits of the
'' high order byte of the first long (see ioCmdMask).  This is always non-zero to indicate
'' that a command is to be performed by the COG routines.  When the command is finished,
'' this is set to zero.  The errorFlag bit is set to one if a NAK was read after a write
'' transfer.  This is the only error reported by these routines.  A read operation and
'' zero-length writes do involve several write transfers for addressing, but the data
'' read transfer has no error checking.  When the command is completed, the device address,
'' byte count, and HUB address are all updated to their values at that time.

'' The pins used for the boot EEPROM I2C bus (at least on Parallax's Demo Board) do not
'' have a pullup on SCL.  This requires that SCL be driven both high and low.  If the bus
'' used is on pins 28 and 29, SCL is actively driven at all times.

'' These EEPROM read/write routines do not provide for waiting for the write to complete
'' nor do they check for paged writes.  All bytes in a multi-byte write must lie within
'' a single EEPROM page since the EEPROM write address counter wraps around at a page
'' boundary.  Similarly, for multi-byte reads, all requested bytes must lie within the
'' same device since the sequential read counter wraps around at the device boundary.

'' Command codes are provided for devices with zero, one, or two address bytes following
'' the device selection byte.  As for all I2C devices, addressing is done using write
'' mode and the device is reselected in read mode after the last address byte.  In the
'' case of ioRead0Cmd, the device is initially selected in read mode.  For 8-bit addresses,
'' the device select code is taken from bits 15-8 of the address value.  For the case
'' without address bytes, the device select code is taken from bits 7-0 of the address value.
'' These device select codes must have their least significant bit set to zero (for write
'' mode) except in the case of ioRead0Cmd where it must be set to one for proper operation.

CON
  i2cDataSet  =   250 * 2 / 25           ' Data setup time -  250ns (in clocks @ 80MHz)
  i2cClkLow1  =  4700 * 2 / 25           ' Clock low time  - 4700ns (in clocks @ 80MHz)
  i2cClkHigh1 =  4000 * 2 / 25           ' Clock high time - 4000ns (in clocks @ 80MHz)
  i2cClkLow4  =  1300 * 2 / 25           ' Clock low time  - 1300ns (in clocks @ 80MHz)
  i2cClkHigh4 =  1000 * 2 / 25           ' Clock high time - 1000ns (in clocks @ 80MHz)

'' Command code and error information for I2C driver
'' (For convenience in using just OS_loaderInit, these are included here.  The "master"
''  copies are considered the ones in OS_loader and these must be kept up-to-date).

  ioReadCmd   = %00000001                      ' Read from EEPROM to HUB RAM (16 bit addresses)
  ioWriteCmd  = %00000010                      ' Write to EEPROM from HUB RAM (16 bit addresses)
  ioRead1Cmd  = %00000011                      ' Read from a device with only 8-bit addresses
  ioWrite1Cmd = %00000100                      ' Write to a device with only 8-bit addresses
  ioRead0Cmd  = %00000101                      ' Read from a device without address bytes
  ioWrite0Cmd = %00000110                      ' Write to a device without address bytes
  ioBootCmd   = %00001000                      ' Read from EEPROM to HUB RAM, then start a
                                               ' new SPIN interpreter in the COG whose ID is
                                               ' supplied in the lower 3 bits of this command
                                               ' This COG is stopped before the read is done.
  ioCmdMask   = %00001111                      ' Used to mask off command bits
  ioLowSpeed  = %01000000                      ' If set, I2C runs at 100KHz rather than 400KHz
  ioWriteErr  = %10000000                      ' An error occurred during an I2C write (NAK)

  ioTestRdy   = ioCmdMask << 24                ' Used to test 1st control long for ready
  ioTestErr   = ioWriteErr << 24               ' Used to test 1st control long for write error

'' Other constants from OS_loader

  screenPtr   = $8000 - 20                     ' Pointer to screen buffer
  ioControl   = $8000 - 48                     ' I/O control block for general use
  i2cBootSCL  = 28                             ' Boot EEPROM SCL pin

VAR
  long cog, control[2]

PUB bootEEPROM(addr) | t                               '' Load and run a new SPIN program
  if not start
    abort                                              ' If no COGs available, abort
  control[1] := 0                                      ' Check for the presence of EEPROM
  control[0] := (ioReadCmd << 24) | (addr & $FFFFFF)
  repeat while control[0] & ioTestRdy                  ' Wait for check to complete and
  if control[0] & ioTestErr                            '  abort if there's an error
    abort
  repeat t from 0 to 7                                 ' Stop all COGs except this one and
    if (t <> cogid) and (t <> (cog-1))                 '  the one with the I2C driver in it
      cogstop(t)
  t := ioBootCmd | cogid                               ' Tell the I2C driver to load 32K
  control[1] := $80000000                              '  into HUB RAM after stopping this
  control[0] := (t << 24) | (addr & $FFFFFF)           '   COG, then starting the SPIN
  cogstop(cogid)                                       '    interpreter on the new program
  
PUB readEEPROM(addr,buffer,count)                      '' Read a block from EEPROM to RAM
  repeat while control[0] & ioTestRdy                  ' Wait for previous I/O to finish
  control[1] := (count << 16) | (buffer & $FFFF)
  control[0] := (ioReadCmd << 24) | (addr & $FFFFFF)
  repeat while control[0] & ioTestRdy                  ' Wait for this to finish
  return (control[0] & ioTestErr) <> 0                 ' Return any error code

PUB writeEEPROM(addr,buffer,count)                     '' Write a block to EEPROM from RAM
  repeat while control[0] & ioTestRdy                  ' Wait for previous I/O to finish
  control[1] := (count << 16) | (buffer & $FFFF)
  control[0] := (ioWriteCmd << 24) | (addr & $FFFFFF)
  repeat while control[0] & ioTestRdy                  ' Wait for this to finish
  return (control[0] & ioTestErr) <> 0                 ' Return any error code

PUB start                                              '' Start the I2C I/O driver (standalone)
  stop                                                 ' Stop a previous copy
  control[0] := 0                                      ' Initialize the control block
  control[1] := 0
  cog := cognew(@i2cEntryPoint,@control) + 1           ' Start a new cog with the I2C driver
  return cog > 0                                       ' Indicate success

PUB stop                                               '' Stop the I2C I/O driver (standalone)
  if cog > 0
    cogstop(cog - 1)

PUB getControl(i)                                      '' Return a long from the control block
  return control[i]                                    ' Check for operation completed first

PUB setControl(i,value)                                '' Set value of a long in the control block
  control[i] := value                                  ' Always set the first long last

PUB getEntryPoint                                      '' Return COG address (for COGINIT)
  return @i2cEntryPoint                                ' This is used by the Operating System
  
DAT
                        org     0
i2cEntryPoint           jmp     #:newOpFetch            ' Go process operation
:newOpWait              mov     i2cTemp,i2cPause
                        add     i2cTemp,cnt             ' Wait 10us before checking
                        waitcnt i2cTemp,#0
:newOpFetch             rdlong  i2cAddr,PAR             ' Fetch control information
                        test    i2cAddr,ReqMask    wz
                if_z    jmp     #:newOpWait             ' Wait for a new operation
                        rdlong  SaveClkFreq,#$0000
                        rdbyte  SaveClkMode,#$0004      ' Save clock frequency and mode
                        mov     i2cCmd,i2cAddr
                        shr     i2cCmd,#24              ' Isolate command code
                        and     i2cCmd,#ioCmdMask
                        movs    :getAction,i2cCmd       ' Get command specific action
                        cmp     i2cCmd,#ioBootCmd  wc   '  bit sequence.  ioBootCmd is
                if_nc   movs    :getAction,#ioReadCmd   '   treated as ioReadCmd here
                        add     :getAction,#ActionTbl
                        nop                             ' Self-modifying code pause
:getAction              mov     Action,0-0
                        mov     i2cClkLow,#i2cClkLow4
                        mov     i2cClkHigh,#i2cClkHigh4
                        test    i2cAddr,speedMask  wz   ' Check for low speed bus
                if_nz   mov     i2cClkLow,#i2cClkLow1
                if_nz   mov     i2cClkHigh,#i2cClkHigh1
                        and     i2cAddr,i2cAddrMask     ' Only need address at this point
                        mov     i2cTemp,PAR
                        add     i2cTemp,#4              ' Now get 2nd long of packet
                        rdlong  i2cCount,i2cTemp
                        mov     i2cBufAdr,i2cCount      ' Byte count
                        shr     i2cCount,#16
                        and     i2cBufAdr,i2cWordMask   ' HUB RAM address of buffer
                        mov     CheckSum,#0
                        cmp     i2cCmd,#ioBootCmd  wc
                if_c    jmp     #:mainCycle
                        mov     i2cTemp,i2cCmd          ' Stop the COG that will be
                        and     i2cTemp,%111            '  used for booting later
                        cogid   i2cCogId                '   unless it's this one
                        cmp     i2cCogId,i2cTemp   wz
                if_ne   cogstop i2cTemp
:mainCycle              mov     i2cTemp,i2cAddr
                        shr     i2cTemp,#18             ' Determine bit masks for
                        and     i2cTemp,#%11110         '  I/O pins for I2C bus
                        mov     i2cSCL,#1
                        shl     i2cSCL,i2cTemp
                        mov     i2cSDA,i2cSCL           ' SDA is next higher pin
                        shl     i2cSDA,#1
                        test    FirstCall,i2cSCL   wz   ' Is this our first call?
                if_nz   andn    FirstCall,i2cSCL        '  if so, do a reset
                if_nz   call    #i2cReset
                        call    #i2cStart               ' Do a start sequence
                        test    Action,#%000000001 wz
                if_z    jmp     #:skipAction0
                        mov     i2cData,i2cAddr         ' Construct a device select
                        shr     i2cData,#15             '  code for EEPROM write mode
                        and     i2cData,#%00001110
                        or      i2cData,#%10100000
                        mov     i2cMask,#%10000000
                        call    #i2cWrite               ' Send device select code
                if_c    jmp     #:doStop                ' Failure if NAK received
:skipAction0            test    Action,#%000000010 wz
                if_z    jmp     #:skipAction1
                        mov     i2cData,i2cAddr         ' First address byte is most
                        shr     i2cData,#8              '  significant byte of address
                        mov     i2cMask,#%10000000
                        call    #i2cWrite               ' Send first address byte
                if_c    jmp     #:doStop                ' Failure if NAK received
:skipAction1            test    Action,#%000000100 wz
                if_z    jmp     #:skipAction2
                        mov     i2cData,i2cAddr         ' Second address byte is least
                        mov     i2cMask,#%10000000      '  significant byte of address
                        call    #i2cWrite               ' Send second address byte
                if_c    jmp     #:doStop                ' Failure if NAK received
:skipAction2            tjz     i2cCount,#:doStop       ' If byte count == 0, we're done
                        test    Action,#%000001000 wz
                if_nz   call    #i2cStart               ' Do a start sequence if readdressing
:doReadWrite            test    Action,#%000010000 wz
                if_nz   rdbyte  i2cData,i2cBufAdr       ' If writing, fetch the data value
                        test    Action,#%000100000 wz
                if_z    jmp     #:skipAction5
                        mov     i2cData,i2cAddr         ' If reading, construct a device select
                        shr     i2cData,#15             '  code for EEPROM read mode
                        and     i2cData,#%00001110
                        or      i2cData,#%10100001
:skipAction5            test    Action,#%001000000 wz
                if_z    jmp     #:skipAction6
                        mov     i2cData,i2cAddr         ' If reading using a single byte address
                        shr     i2cData,#8              '  construct a device select code for
                        or      i2cData,#%00000001      '   read mode given one for write mode
:skipAction6            test    Action,#%010000000 wz
                if_z    jmp     #:skipAction7
                        mov     i2cMask,#%10000000      ' Either readdress device for reading
                        call    #i2cWrite               '  or write a data value at this point
                if_c    jmp     #:doStop                ' Failure if NAK received
:skipAction7            test    Action,#%100000000 wz
                if_z    jmp     #:skipAction8
                        cmp     i2cCount,#2        wc   ' Carry true if this is the last byte
                        mov     i2cMask,#%10000000
                        call    #i2cRead
                        or      i2cZero,#0      nr,wc   ' Force carry false to show success
                        wrbyte  i2cData,i2cBufAdr
                        add     CheckSum,i2cData        ' Accumulate checksum for ioBootCmd
                        andn    Action,#%011100000      ' No readdressing on subsequent reads
:skipAction8            add     i2cAddr,#1
                        add     i2cBufAdr,#1            ' On to next byte to be transferred
                        djnz    i2cCount,#:doReadWrite  ' Repeat for number of bytes requested
:doStop                 call    #i2cStop
                if_c    or      i2cAddr,errorFlag       ' Carry true indicates error
                if_c    jmp     #:allDoneCheck
                        cmp     i2cCmd,#ioBootCmd  wc   ' See if booting after transfer
                if_c    jmp     #:allDone
                        and     CheckSum,#$FF      wz   ' Checksum must be zero
                if_nz   jmp     #:stopThisCOG
                        rdlong  i2cTemp,#$0000          ' Is clock frequency the same?
                        cmp     i2cTemp,SaveClkFreq wz
                        rdbyte  i2cTemp,#$0004          ' Is clock mode the same as well?
                if_ne   jmp     #:changeClock
                        cmp     i2cTemp,SaveClkMode wz  ' If both same, just go start COG
                if_e    jmp     #:justStartUp
:changeClock            and     i2cTemp,#$F8            ' If xtal/pll enabled, start up now
                        clkset  i2cTemp                 '  while remaining in current mode
                        mov     i2cTemp,time_xtal
:startupDelay           djnz    i2cTemp,#:startupDelay  ' Allow 20ms@20MHz for xtal/pll to settle
                        rdbyte  i2cTemp,#$0004          ' Switch to selected clock
                        clkset  i2cTemp
:justStartUp            mov     i2cTemp,i2cCmd          ' Use the COG supplied as the caller's
                        and     i2cTemp,#%111           '  to start up the SPIN interpreter
                        or      i2cTemp,interpreter     '   (It can be the same as this COG)
                        coginit i2cTemp            wc   ' Reboot the cog with SPIN interpreter
                if_c    jmp     #:stopThisCOG           ' If not started, stop this loader COG
                        cmp     ioControlAdr,PAR        ' If using OS's I/O Control Block, go
                if_e    jmp     #:allDone               '  update contents and wait for request
:stopThisCOG            cogid   i2cTemp
                        cogstop i2cTemp                 ' Stop this COG if just loaded memory
:allDoneCheck           cmp     i2cCmd,#ioBootCmd  wc   ' See if intended booting after transfer
                if_nc   jmp     #:stopThisCOG           '  if so, an error occurred, go stop COG
:allDone                and     i2cBufAdr,i2cWordMask   ' Copy updated information
                        shl     i2cCount,#16            '  back to control packet
                        or      i2cCount,i2cBufAdr
                        mov     i2cTemp,PAR
                        add     i2cTemp,#4
                        wrlong  i2cCount,i2cTemp
                        wrlong  i2cAddr,PAR             ' Indicate operation is done
                        jmp     #:newOpWait             '  and go wait for a new one

'' This action table contains bit sequences for controlling device addressing and read/write
'' mode selection for each of the commands possible.  From LSB to MSB, the actions are:
'' 0 - Write the EEPROM device select code for write mode
'' 1 - Write the MSB device address or (for ioRead1Cmd/ioWrite1Cmd) a device select code
'' 2 - Write the LSB device address or (for ioRead0Cmd/ioWrite0Cmd) a device select code
'' 3 - Output a Start Sequence prior to reselecting in read mode
'' 4 - Fetch a data value for writing
'' 5 - Construct an EEPROM device select code for read mode
'' 6 - Construct a read mode device select code from the MSB of the 16 bit device address
'' 7 - Write the data value or read mode device select code
'' 8 - Read a byte of data from the device and store it

ActionTbl               long    %000000000              ' Command not used (indicates done)
                        long    %110101111, %010010111  ' Read/Write with 2 bytes of addressing
                        long    %111001110, %010010110  ' Read/Write with 1 byte of addressing
                        long    %100000100, %010010100  ' Read/Write data only
                        long    %000000000              ' Command not used
Action                  long    0
i2cCogId                long    0
i2cCmd                  long    0
i2cZero                 long    0                       ' Zero (or False) value
i2cOne                  long    1                       ' One value
i2cTrue                 long    -1                      ' True (-1) value
FirstCall               long    -1
i2cTemp                 long    0
i2cHighMask             long    $00FF00
i2cCount                long    0
i2cBufAdr               long    0
i2cAddr                 long    0
i2cWordMask             long    $0000FFFF
i2cAddrMask             long    $00FFFFFF
ReqMask                 long    $3F000000               ' Request new operation if <> 0
errorFlag               long    $80000000               ' NAK received during write cycle
speedMask               long    $40000000               ' One if 100KHz bus, zero if 400KHz
i2cPause                long    10000 * 2 / 25          ' Clock cycles for 10us @ 80 MHz
time_xtal               long    20 * 20000 / 4 / 1      ' 20ms (@20MHz, 1 inst/loop)
interpreter             long    ($0004 << 16) | ($F004 << 2) | %0000
i2cClkLow               long    0                       ' Minimum clock low time (ticks)
i2cClkHigh              long    0                       ' Minimum clock high time (ticks)
SaveClkFreq             long    0                       ' Initial clock frequency ($0000)
SaveClkMode             long    0                       ' Initial clock mode value ($0004)
CheckSum                long    0                       ' Checksum of bytes for ioBootCmd
ioControlAdr            long    ioControl               ' Address of OS I/O Control Block

'' Low level I2C routines.  These are designed to work either with a standard I2C bus
'' (with pullups on both SCL and SDA) or the Propellor Demo Board (with a pullup only
'' on SDA).  Timing can be set by the caller to 100KHz, 400KHz, or 1MHz.

'' Do I2C Reset Sequence.  Clock up to 9 cycles.  Look for SDA high while SCL
'' is high.  Device should respond to next Start Sequence.

i2cReset                andn    dira,i2cSDA             ' Pullup drive SDA high
                        mov     i2cBitCnt,#9            ' Number of clock cycles
                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Allow for minimum SCL low
:i2cResetClk            waitcnt i2cTime,i2cClkHigh
                        test    i2cBootSCLm,i2cSCL wz   ' Check for boot I2C bus
              if_nz     or      outa,i2cSCL             ' Active drive SCL high
              if_nz     or      dira,i2cSCL
              if_z      andn    dira,i2cSCL             ' Pullup drive SCL high
                        waitcnt i2cTime,i2cClkLow       ' Allow minimum SCL high
                        test    i2cSDA,ina         wz   ' See if SDA is high
                        andn    outa,i2cSCL             ' Active drive SCL low
                        or      dira,i2cSCL            
              if_z      djnz    i2cBitCnt,#:i2cResetClk ' Stop when SDA was high
i2cReset_ret            ret                             ' Should be ready for Start      

'' Do I2C Start Sequence.  This assumes that SDA is a floating input and
'' SCL is also floating, but may have to be actively driven high and low.
'' The start sequence is where SDA goes from HIGH to LOW while SCL is HIGH.

i2cStart                andn    dira,i2cSDA             ' Pullup drive SDA high
                        andn    outa,i2cSDA             ' SDA set to drive low
                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Allow for bus free time
                        waitcnt i2cTime,i2cClkHigh
                        test    i2cBootSCLm,i2cSCL wz   ' Check for boot I2C bus
              if_nz     or      outa,i2cSCL             ' Active drive SCL high
              if_nz     or      dira,i2cSCL
              if_z      andn    dira,i2cSCL             ' Pullup drive SCL high
                        waitcnt i2cTime,i2cClkHigh      ' Allow for start setup time
                        or      dira,i2cSDA             ' Active drive SDA low
                        waitcnt i2cTime,#0
                        andn    outa,i2cSCL             ' Active drive SCL low
                        or      dira,i2cSCL
i2cStart_ret            ret                             

'' Do I2C Stop Sequence.  This assumes that SCL is low and SDA is indeterminant.
'' The stop sequence is where SDA goes from LOW to HIGH while SCL is HIGH.
'' i2cStart must have been called prior to calling this routine for initialization.
'' The state of the (c) flag is maintained so a write error can be reported.

i2cStop                 or      dira,i2cSDA             ' Active drive SDA low
                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Wait for minimum clock low
                        waitcnt i2cTime,i2cClkLow
                        test    i2cBootSCLm,i2cSCL wz   ' Check for boot I2C bus
              if_nz     or      outa,i2cSCL             ' Active drive SCL high
              if_nz     or      dira,i2cSCL
              if_z      andn    dira,i2cSCL             ' Pullup drive SCL high
                        waitcnt i2cTime,#0              ' Wait for minimum setup time
                        andn    dira,i2cSDA             ' Pullup drive SDA high
i2cStop_ret             ret                             ' Leave SCL and SDA high

'' Write I2C data.  This assumes that i2cStart has been called and that SCL is low,
'' SDA is indeterminant. The (c) flag will be set on exit from ACK/NAK with ACK == false
'' and NAK == true. Bytes are handled in "little-endian" order so these routines can be
'' used with words or longs although the bits are in msb..lsb order.

i2cWrite                mov     i2cBitCnt,#8
                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Wait for minimum SCL low
:i2cWriteBit            waitcnt i2cTime,#i2cDataSet
                        test    i2cData,i2cMask    wz
              if_z      or      dira,i2cSDA             ' Copy data bit to SDA
              if_nz     andn    dira,i2cSDA
                        waitcnt i2cTime,i2cClkHigh      ' Wait for minimum setup time
                        test    i2cBootSCLm,i2cSCL wz   ' Check for boot I2C bus
              if_nz     or      outa,i2cSCL             ' Active drive SCL high
              if_nz     or      dira,i2cSCL
              if_z      andn    dira,i2cSCL             ' Pullup drive SCL high
                        waitcnt i2cTime,i2cClkLow
                        andn    outa,i2cSCL             ' Active drive SCL low
                        or      dira,i2cSCL
                        ror     i2cMask,#1              ' Go do next bit if not done
                        djnz    i2cBitCnt,#:i2cWriteBit
                        andn    dira,i2cSDA             ' Switch SDA to input and
                        waitcnt i2cTime,i2cClkHigh      '  wait for minimum SCL low
                        test    i2cBootSCLm,i2cSCL wz   ' Check for boot I2C bus
              if_nz     or      outa,i2cSCL             ' Active drive SCL high
              if_nz     or      dira,i2cSCL
              if_z      andn    dira,i2cSCL             ' Pullup drive SCL high
                        waitcnt i2cTime,#0              ' Wait for minimum high time
                        test    i2cSDA,ina         wc   ' Sample SDA (ACK/NAK) then
                        andn    outa,i2cSCL             '  active drive SCL low
                        or      dira,i2cSCL
                        rol     i2cMask,#16             ' Prepare for multibyte write
i2cWrite_ret            ret

'' Read I2C data.  This assumes that i2cStart has been called and that SCL is low,
'' SDA is indeterminant.  ACK/NAK will be copied from the (c) flag on entry with
'' ACK == low and NAK == high.  Bytes are handled in "little-endian" order so these
'' routines can be used with words or longs although the bits are in msb..lsb order.

i2cRead                 mov     i2cBitCnt,#8
                        andn    dira,i2cSDA             ' Make sure SDA is set to input
                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Wait for minimum SCL low
:i2cReadBit             waitcnt i2cTime,i2cClkHigh
                        test    i2cBootSCLm,i2cSCL wz   ' Check for boot I2C bus
              if_nz     or      outa,i2cSCL             ' Active drive SCL high
              if_nz     or      dira,i2cSCL
              if_z      andn    dira,i2cSCL             ' Pullup drive SCL high
                        waitcnt i2cTime,i2cClkLow       ' Wait for minimum clock high
                        test    i2cSDA,ina         wz   ' Sample SDA for NAK/ACK
                        andn    outa,i2cSCL             ' Active drive SCL low
                        or      dira,i2cSCL
              if_nz     or      i2cData,i2cMask         ' Accumulate data bits
              if_z      andn    i2cData,i2cMask
                        ror     i2cMask,#1              ' Shift the bit mask and
                        djnz    i2cBitCnt,#:i2cReadBit  '  continue until done
                        waitcnt i2cTime,#i2cDataSet     ' Wait for end of SCL low
              if_c      andn    dira,i2cSDA             ' Copy the ACK/NAK bit to SDA
              if_nc     or      dira,i2cSDA
                        waitcnt i2cTime,i2cClkHigh      ' Wait for minimum setup time
                        test    i2cBootSCLm,i2cSCL wz   ' Check for boot I2C bus
              if_nz     or      outa,i2cSCL             ' Active drive SCL high
              if_nz     or      dira,i2cSCL
              if_z      andn    dira,i2cSCL             ' Pullup drive SCL high
                        waitcnt i2cTime,#0              ' Wait for minimum clock high
                        andn    outa,i2cSCL             ' Active drive SCL low
                        or      dira,i2cSCL
                        rol     i2cMask,#16             ' Prepare for multibyte read
i2cRead_ret             ret

'' Local variables for low level I2C routines

i2cSCL                  long    0                       ' Bit mask for SCL
i2cSDA                  long    0                       ' Bit mask for SDA
i2cTime                 long    0                       ' Used for timekeeping
i2cData                 long    0                       ' Data to be transmitted / received
i2cMask                 long    0                       ' Bit mask for bit to be tx / rx
i2cBitCnt               long    0                       ' Number of bits to tx / rx
i2cBootSCLm             long    |<i2cBootSCL            ' Bit mask for pin 28 SCL use

'' Debugging routines that write directly to the display buffer

debugHexPart            mov     debugShift,#0
                        mov     debugChr,#0
:debugHexLoop           rol     debugVal,#4
                        mov     debugDig,debugVal
                        and     debugDig,#%1111
                        add     debugDig,#"0"
                        cmp     debugDig,#"9"+1    wc
              if_nc     add     debugDig,#"A"-("9"+1)
                        shl     debugDig,debugShift
                        or      debugChr,debugDig
                        add     debugShift,#8
                        djnz    debugCnt,#:debugHexLoop
                        call    #debugOut
debugHexPart_ret        ret

debugHexByte            mov     debugCnt,#2
                        shl     debugVal,#24
                        call    #debugHexPart
debugHexByte_ret        ret

debugHexWord            mov     debugCnt,#4
                        shl     debugVal,#16
                        call    #debugHexPart
debugHexWord_ret        ret

debugHexLong            mov     debugHexPart_ret,debugHexLong_ret
                        mov     debugCnt,#4
                        call    #debugHexPart
                        mov     debugCnt,#4
                        call    #debugHexPart
debugHexLong_ret        ret

debugInit               rdword  debugDsp,screenPtrAdr    ' Write directly to the screen
debugInit_ret           ret                              

debugOut                or      debugChr,#0        wz    ' Stop when debugChr all zero
              if_z      jmp     #debugOut_ret
:debugNext              wrbyte  debugChr,debugDsp        ' Write char & move to next
                        add     debugDsp,#1             
                        shr     debugChr,#8        wz    ' Order is LSB..MSB          
              if_nz     jmp     #:debugNext
debugOut_ret            ret

screenPtrAdr            long    screenPtr                ' Address of screen pointer in HUB RAM
debugDsp                long    0                        ' Pointer to display memory
debugChr                long    0                        ' Up to 4 characters to output
debugVal                long    0
debugDig                long    0
debugShift              long    0
debugCnt                long    0
                        fit